<?php
//////////////////////////////////////////////////////////////////////
// Generic Tree Utility (GTU)
// v3.0.0, January 31, 2003
// by:  Mike Agar
//////////////////////////////////////////////////////////////////////
// This code is released under GPL refer to the included license.txt
//////////////////////////////////////////////////////////////////////

class agar_gtu {
    var $gtu_debug;
    var $gtu_table_name;

//STATE MACHINE VARIABLES
    var $icons;                            // holds the icon img info for each type for each state
    var $visible;                        // flag to indicate visibility for each state (default=1)
    var $url;                                // url to apply to the node handle for each state
    var $target;                        // where to open the page for each type for each state
    var $action;                        // the action array (controls for an item)
    var $c_root;              // the current root and state
    var $c_rid;                  // the current root id
    var $get_text_func;       // call back functions for each type for each state
    var $show_child_count;     // display attribute: show the " (..)" for each parent?
    var $jump;                             // display attribute: use "#" notation in tree
    var $all_in_context;        //

// new vars
    var $widgetPath="MyHeadlines/images/icon/";
    var $useWidgets=false; 

/*********************************************************************************************
*** THIS MUST BE CALLED BEFORE ANY OTHER GTU FUNCTION !!!  ***********************************
**********************************************************************************************
    Required: gtu table handles (pretty important)
    Optional: restore the State Machine from the cache table

*/
    function GTU_INIT($tableNamePrefix="AGAR_gtu"){
        $this->gtu_table_name=$tableNamePrefix; //MUST BE FIRST!!!
        $found = false;
        //check to see if this table exisits... if not, create it.
        $result = mysql_query("show tables");
        while(list($handle) = mysql_fetch_row($result)){
            if(stristr($handle,$this->gtu_table_name)){
                $found|=true;
            }
        }
        if($found){//new columns added Jan 31, 2003...  One day we'll clean this up...
            $result = mysql_query("show columns from ".$this->gtu_table_name);
            if(mysql_num_rows($result)<6){
                $this->updateGTUTables();
            }
        }else{
            $this->createGTUTables();
        }
    }



/*********************************************************************************************
*** State Machine Functions
*********************************************************************************************/

/*********************************************************************************************
GTU_SELECT():
~~~~~~~~~~~~
This configures the GTU State Machine. The user must call this function immediately after the 
initialization. It store/sets the current root, and the parameters for this tree.  Remember
that there may be more that one tree in the GTU, and each tree may have unique parameters
to define the state of that tree.

$rootName                  = registered handle of a tree, used to set current root
$state                     = user defined tag to identify cached state machine settings
$show_count                 = this is a visual attribute used to indicate the number of children
$jump                            = jump to the position of the drilled item (use "#" notation)
$allInContext             = turn on/off Expand the tree fully (everyhting is in the context).
$debug                         = turn on/off (1/0) debugging info.


*********************************************************************************************/
    function GTU_SELECT(    $rootName,
                            $state="default",
                            $show_count=1,
                            $jmp=1,
                            $allInContext=0,
                            $debug=0 ){
        $this->setCurrentRoot($rootName,$state);
        $this->gtu_debug=$debug;
        if($debug){
            echo "GTU_SELECT(): Debug is ON = $this->gtu_debug<br>\n";
        }
        $this->show_child_count=$show_count;
        $this->jump=$jmp;
        $this->all_in_context=$allInContext;
    }

    function setGetTextCallBack($getTextCallBack){
    /***************************************************************************************************
        the GTU has no outside knowledge.  It doesnot know what is stored in its tables.
        It is a logical construct used to represent a directory tree.  The catch is that
        the internal display functions need to have a way for asking for a handle of a given
        item.  The callback function must be provided to GTU for EVERY User Defined State:
        
            callBack($eid,$tid){}

        Where $eid is the external id, and $tid is the external type.  This function must return
        to the GTU a string that will be displayed, a handle, to represent the unique item-type 
        pair of eid-tid.

        -> IF YOU CALLED "GTU_SELECT" TO START A NEW STATE MACHINE YOU MUST REGISTER A  <-
        -> CALL BACK FUNCTION !!!                                                       <-

    */
        $this->get_text_func=$getTextCallBack;
        $this->saveStateMachine();//store the state machine in the database
    }

    function setCurrentRoot($rootName,$state){
        $rid = $this->getRoot($rootName);
        if($rid<=0){
            $rid = $this->addRoot($rootName);
        }
        $this->c_root = "$rootName"."_".$state;
        $this->c_rid = $rid;
        $this->restoreStateMachine();
    }

    function setTypeID($tid){//inits defaults for type
        //init default values for this state
        if($this->gtu_debug)echo "%%% setTypeID(): tid=$tid<br>\n";
        $this->icons[$tid]="";
        $this->visible[$tid]=1;
        $this->url[$tid]=0;
        $this->target[$tid]=0;
        $this->action[$tid]=array();
        $this->saveStateMachine();//store the state machine in the database
    }

    function setIcon($tid,$icon){
        $this->icons[$tid]=$icon;
        $this->saveStateMachine();//store the state machine in the database
    }

    function setVisible($tid,$vis=1){
        $this->visible[$tid]=$vis;
        $this->saveStateMachine();//store the state machine in the database
    }

    function setURL($tid,$url){
        $this->url[$tid]=$url;
        $this->saveStateMachine();//store the state machine in the database
    }

    function setTarget($tid,$target){
        $this->target[$tid]=$target;
        $this->saveStateMachine();//store the state machine in the database
    }

    function addAction($tid,$text,$url,$target=""){
    //For each type in each state you may add many actions.
    //action is an array of serialized arrays
        $act=array();
        $act[url]=$url;
        $act[text]=$text;
        $act[target]=$target;
        $this->action[$tid][] = serialize($act);
        $this->saveStateMachine();
    }

/*********************************************************************************************
*** Init/Add GTU Functions
*********************************************************************************************/
    function addRoot($desc){ //alias for naming conventions
        return $this->newRoot($desc);
    }
    
    function newRoot($desc){
    /******************************************************************
    returns 0  if description already exists
    returns -1 if other failure
    returns rootID if successful
    ******************************************************************/

        $test=$this->getRoot($desc);
        if($test==-1){
            $result = mysql_query("insert into $this->gtu_table_name values(NULL,'99','99','99','99',1)");
            $gid = mysql_insert_id();
            $result = mysql_query("update $this->gtu_table_name set pid='$gid', eid='$gid', tid='$gid', rid='$gid' where gid='$gid'");   
            $result = mysql_query("insert into $this->gtu_table_name"."_reg values($gid, '$desc')");
            return $this->getRoot($desc);
        }else{
            if($test>0){
                return 0;
            }else{
                return -1;
            }
        }
    }

    function addNode($pid,$eid,$tid,$weight=1){
        //assumes you have already selected the tree using GTU_SELECT()
       $gid = $this->nodeExists($pid,$eid,$tid);
         if(!$gid){
           $result = mysql_query("insert into $this->gtu_table_name values(NULL,'$pid','$eid','$tid','".$this->c_rid."',$weight)");
             $gid = mysql_insert_id();
         }else{
            //make sure weight is preserved (incase it changes)
            $this->modNodeWeight($gid,$weight);
         }
         return $gid;  
    }



/*********************************************************************************************
*** Maintenance GTU Functions
*********************************************************************************************/

//----> ROOT Functions <----//
    function getRootGID($desc){ //alias for naming conventions
        return $this->getRoot($desc);
    }

    function getRoot($desc){
    /*******************************************
     returns -1 if description doesn't exist
     returns -2 if root is corrupt
     returns rootID if successful
    ********************************************/

       $result = @mysql_query("select gid from $this->gtu_table_name"."_reg where handle='$desc'");
       if(!$result)return -3;
         if (!mysql_num_rows($result)){
             return -1;
       }else{
             list($gid) = mysql_fetch_row($result);
        $result = mysql_query("select gid, pid, eid, tid from $this->gtu_table_name where gid='$gid'");
        list($temp_gid,$pid,$eid,$tid) = mysql_fetch_row($result);
        if(($temp_gid != $gid)
            ||($pid != $gid)
            ||($eid != $gid)
            ||($tid != $gid)){
          /****** Failed sanity check ******/
                 return -2;   
        }
        return $gid;
       }
    }

    function isRoot($gid){
         $result = mysql_query("select * from $this->gtu_table_name"."_reg where gid='$gid'");
         return mysql_num_rows($result);
    }

    function delRoot($desc){
      $rt = $this->getRoot($desc);
      if($rt>=0){
          $result = @mysql_query("delete from $this->gtu_table_name where rid = $rt");
          $result = @mysql_query("delete from $this->gtu_table_name"."_reg where gid = $rt");
            return 0;
      }else{
        return -1;
      }
    }

//----> NODE Functions <----//
    function copyNode($fromGID,$toPID,$andChildren=0,$eid=0,$tid=0){
        if(!$eid)$eid = $this->getNodeEID($fromGID);
        if(!$tid)$tid = $this->getNodeTID($fromGID);
        $w = $this->getNodeWeight($fromGID);
        $newPID = $this->addNode($toPID,$eid,$tid,$w);
        if($andChildren){
            $tree = $this->listNodes($fromGID);
            if(is_array($tree)){
                //has children to be copied...
                while(list($ogid,$opid,$oeid,$otid) = @current($tree)){
                    $dumpvar = $this->copyNode($ogid,$newPID,$andChildren,$oeid,$otid);//Copy children
                    next($tree);
                }
            }
        }
        return $newPID;
    }

    function moveNode($fromGID,$toPID){
        $newGID = $this->copyNode($fromGID,$toPID,1);
        $ic = $this->isChild($fromGID,$toPID);
        if(!$ic){
            $err = $this->delNode($fromGID,1);
        }else{
            if($this->gtu_debug)echo "moveNode() - Sorry but you can't move this branch.  (isChild()==$ic)<br>\n";
        }
        return $newGID;
    }



    function modNode($gid, $what, $where){
        $result = mysql_query("update $this->gtu_table_name set $what='$where' where gid='$gid'");
        if(@mysql_errno()){
            echo "There was a mysql error during this operation.<br>";
        }
    }

    function modNodeTID($gid,$TID){
        $this->modNode($gid,"tid",$TID);
    }

    function modNodeEID($gid,$EID){
        $this->modNode($gid,"eid",$EID);
    }

    function modNodePID($gid,$PID){
        $this->modNode($gid,"pid",$PID);
    }

    function modNodeRID($gid,$RID){
        $this->modNode($gid,"rid",$RID);
    }

    function modNodeWeight($gid,$w){
        $result = mysql_query("update $this->gtu_table_name set weight=$w where gid='$gid'");
        if(@mysql_errno()){
            echo "There was a mysql error during this operation.<br>";
        }
    }


    function delNode($gid,$recursive=0){
      if($this->isRoot($gid)){
            return 1; //can't delete root from here
        }
        $children = $this->listNodes($gid);
        if (is_array($children)){
            if ($recursive){
                while(list($nid,$pid,$eid,$tid) = current($children)){
                    if($pid==$gid){//sanity check ;-)
                        $err = $this->delNode($nid,$recursive);
                        if ($err) {
                            return $err;
                        }
                    }
                    next($children);
                }
            }else{
              //has children but didnot say to drop them... return error
                return 1;
            }
        }
      //delete the node
        $result = mysql_query("delete from $this->gtu_table_name where gid='$gid'");
        return 0;

    }

    function delAllOfType($tid){
        //only do node in the same tree ($rid)
        $query = "delete from $this->gtu_table_name where tid='$tid' and rid='".$this->c_rid."'";
        $result = mysql_query($query);
    }

    function delAllMatchingNodes($gid,$deleid,$deltid){
        $tree = $this->listNodes($gid);
        while(list($nid,$pid,$eid,$tid) = @current($tree)){
            if(($eid==$deleid)&&($tid==$deltid)){
                //found a candidate to smack...
                $this->delNode($nid,1);
            }else{
                $children = $this->listNodes($nid);
                if(is_array($children)&&!$this->isRoot($nid)){
                    $this->delAllMatchingNodes($nid,$deleid,$deltid);
                }
            }
            next($tree);
            unset($children);
        }
    }

/*********************************************************************************************
*** Query GTU Functions
*********************************************************************************************/

    function getTreeName($gid){
    /************************************************************************
    finds the tree handle (root handle) that $gid is 
    a member of
    ************************************************************************/
        $rid = $this->getNodeRID($gid);
        $result = mysql_query("select handle from $this->gtu_table_name"."_reg where gid='$rid'");
        list($handle) = mysql_fetch_row($result);
        return $handle;
    }

    function getNode($gid, $what){
        $out=0;
        $result = mysql_query("select $what from $this->gtu_table_name where gid='$gid'");
        list($out) = mysql_fetch_row($result);
        if(@mysql_errno()){
            echo "There was a mysql error during this operation.<br>";
        }
        return $out;
    }


    function getNodePID($gid){
        return $this->getNode($gid,'pid');
    }

    function getNodeEID($gid){
        return $this->getNode($gid,'eid');
    }

    function getNodeTID($gid){
        return $this->getNode($gid,'tid');
    }

    function getNodeRID($gid){
        return $this->getNode($gid,'rid');
    }

    function getNodeWeight($gid){
        return $this->getNode($gid,'weight');
    }

    function nodeExists($pid,$eid,$tid){ //returns node id
        $result = @mysql_query("select gid from $this->gtu_table_name where pid='$pid' and eid='$eid' and tid='$tid'");
        $gid=0;
        if (mysql_num_rows($result)>0){
            list($gid) = mysql_fetch_row($result);
        }
        return $gid;
    }


    function listAllOfType($tid){
        //only do nodes in the same tree ($rid)
        $nodes = array();
        $query = "select eid from $this->gtu_table_name where tid=".$tid." and rid='".$this->c_rid."'";
        $result = mysql_query($query);
        if(!$result)return $nodes;
        $term = mysql_num_rows($result);
        for($loop=0;$loop<$term;$loop++){
            list($eid) = mysql_fetch_row($result);
            $nodes[$eid]=$eid;
        }
        return array_values($nodes);
    }

    function listNodes($gid){//list children of gid
            $result = mysql_query("select gid, pid, eid, tid from $this->gtu_table_name where pid='$gid' order by tid, weight asc");
          if(!$result)return;
            $term = mysql_num_rows($result);
            for($loop=0;$loop<$term;$loop++){
                $nodes[]=mysql_fetch_array($result);
            }
            return $nodes;
    }

    function listTypeOfChild($pid,$type){
    /**** Returns EID of all children matching type ************/
        $children = array();
        $nodes = $this->listNodes($pid);
        while(list($nid,$pid,$eid,$tid) = @each($nodes)){
            if($tid==$type)$children[]=$eid;
        }
        return $children;
    }


    function findNodes($eid,$tid){
        //returns an array of GIDs that are all nodes matching eid & tid
        //UPDATE Jan 31, 2003:  only do node in the same tree ($rid)
        $query = "select gid from $this->gtu_table_name where eid='$eid' and tid='$tid' and rid='".$this->c_rid."'";
        $result = mysql_query($query);
        $nodes = array();
        while(list($gid) = @mysql_fetch_row($result)){
            $nodes[]=$gid;
        }
        return $nodes;
    }


    function isChild($parent,$child){//is this child a member of this parent's family tree?
        $tree = $this->listNodes($parent);
        while(list($nid,$pid,$eid,$tid) = @current($tree)){
            if($nid==$child) return 1;
            $kids = $this->listNodes($nid);
            if(is_array($kids)&&!$this->isRoot($nid)){
                if($this->isChild($nid,$child))return 1;
            }
            next($tree);
            unset($kids);
        }
        return 0;
    }

  function isDrillable($gid){
        //does this node have children?
        $result = mysql_query("select gid from $this->gtu_table_name where pid='$gid'");
        if(!$result)return 0 ;
        if(mysql_num_rows($result)>0)return 1;
        return 0;
  }



/**********************************************************************
*** Display routines
**********************************************************************/
function showListOfNodes($tid, $url, $context, $commands=0){
//NOTE:  Every url will be suffixed by &gid=xxx&pid=xxx&eid=xxx&tid=xxx&context=xxx
    $output = "";
    $rid = $this->c_rid;
    $query = "select distinct eid from ".$this->gtu_table_name." where tid='$tid' and rid='$rid'";
    $result = mysql_query($query);
    if($result){
        if(mysql_num_rows($result)>0){
            //something to see here...
            $output .= "<ul>";
            while(list($external_id) = @mysql_fetch_row($result)){
                //find the node dimension (gid/pid/eid/tid)
                $query = "select gid,pid,eid,tid from ".$this->gtu_table_name." where tid='$tid' and eid='$external_id' and rid='$rid'";
                $r2 = mysql_query($query);
                list($igid,$ipid,$ieid,$itid) = @mysql_fetch_row($r2); //there may be many, but we only need the first
                //woohoo We've got one!!!
                $suf_url = "&gid=$igid"; 
                $suf_url .= "&pid=$ipid"; 
                $suf_url .= "&eid=$ieid"; 
                $suf_url .= "&tid=$itid"; 
                $suf_url .= "&context=$context"; 
                $out_url = $url.$suf_url;
                //use the callback function to figure out the text...
                $func = $this->get_text_func;
                $item = $func($ieid,$itid);
                if(strlen($item)>0){
                    $output .= "<li><a href=$out_url>$item</a>\n";
                }else{
                    $output .= "<li><a href=$out_url> ??? </a>\n";
                }
                if(is_array($commands)){
                    //split key pairs
           			$keys = array_keys($commands);
        			$end = count($keys);
        			for($loop=0;$loop<$end;$loop++){
        			    $output.="[ <a href=".$commands[$keys[$loop]].$suf_url.">".$keys[$loop]."</a> ]";
        			}

                }
            }
            $output .= "</ul>";
        }else{
            //nothing to see here...
            $output .= "<ul>-</ul>";
        }
    }
    $output .= "<p>\n";
    return $output;
}

function drawContextTree($gid=0,$depth=0,$context='',$firstTime=1){
/***********************************************************************************
This could be a case of being too clever for my own good.  I had a vision
of making a function that was so highly flexible that it would be
used over and over again to draw the tree AND trip it out with all of the
functions for the items in the tree.  This will provide different views
of the same data.

Lets deal with it from an API perspective:

$gid,$depth,$context - These allow this function to be recursive.  I won't touch
                       on that here, but it is proven technology from hcu.php
                                       (MyHeadlines v2.0, June 2001)
$firstTime = 0 - this indicates if this is the outer call to the recursive
                 function.  (sets up and tears down the <table>)

GLOBALS USED:
$visible - This is an array of the visible "types" in the tree.  Display
           only the visible type items.  (ex: ignore all others in the tree)
$url - This is the hyperlink to be applied to EACH item of this type.
       Represented by an array (array[$cha_type]="http://..../...?cmd=...")
             This way each visible type may have a link associated with it.
$action - Another array...  this is the trailing commands beside the item.
          I am still thinking about this one...
$icons - this is an array that references what image to show for each type.

Changes: $depth now is an array indicating which icons to use to prefix the 
         the line...

***********************************************************************************/
    $output = "";
    if($this->gtu_debug)echo "%%% drawContextTree() - START<br>\n";
    if($gid==0)$gid = $this->c_rid;
    if($firstTime)$output .= "<table cellpadding=0 cellspacing=0 ><tr><td nowrap>\n";
    $tree = $this->listNodes($gid);
    $cnt = count($tree);
    if($this->gtu_debug)echo "%%% drawContextTree() - Node Count = $cnt ($gid)<br>\n";
    if(!is_array($depth))$depth=array();
    $lastRow=false;
    while(list($nid,$pid,$eid,$tid) = @current($tree)){
        $cnt--;
        if($cnt<=0)$lastRow=true;
        if((!$this->isRoot($nid))&&($this->isVisible($tid))){
            if($this->gtu_debug)echo "%%% drawContextTree() - Yes it's visible and not a root ($nid,$pid,$eid,$tid)<br>\n";
            $output .= $this->indent($depth);
            if($this->gtu_debug)echo "%%% drawContextTree() - ($nid,$pid,$eid,$tid)<br>\n";
            //build url suffix
            if($this->isDrillable($nid)){
                //is toggle the context
                $newContext=$this->toggleContext($context,$nid);
                if($this->jump){
                    $newContext .= "#".$nid;
                }
            }else{
                if($this->gtu_debug)echo "%%% drawContextTree() - NOT Drillable<br>\n";
                $newContext=$context;
            }
            $suffix = "&gid=".$nid."&pid=".$pid."&eid=".$eid."&tid=".$tid."&context=".$newContext;
            //show Item
            if($this->url[$tid]){
                $output .= "<a name=\"$nid\"></a>";  //if drilled, then we may want to jump to the drilled location :)
                $output .= "<a href=\"".$this->url[$tid].$suffix."\"";
                if($this->target[$tid]){
                    $output .= $this->target[$tid];
                }
                $output .= ">";
            }

            //apply widget
            $output .= $this->widget($this->isDrillable($nid),$this->inContext($context,$nid),$lastRow);
            //show Icon
            $icon = $this->icons[$tid];
            $output .= "$icon ";

            $func = $this->get_text_func;
            $item = $func($eid,$tid);

            $output .= "$item";
            $children = "";
            if($this->isDrillable($nid)){
                //is toggle the context
                $children = $this->listNodes($nid);
                if($this->show_child_count){
                    $ccnt = count($children);
                    $output .= " ($ccnt)";
                }
            }
            if($this->url[$tid]){
                $output .= "</a>\n";
            }
            //now if there are any controls...
            $output .= "</td><td>\n";
            $act  = $this->action[$tid];
            $acnt = count($act);
            if($acnt>0){
                $suffix = "&gid=".$nid."&pid=".$pid."&eid=".$eid."&tid=".$tid."&context=".$context;
                $output .= "<nobr>[";
                $output .= $this->showActions($suffix,$tid);
                $output .= " ]</nobr>\n";
            }
            $output .= "</td></tr>\n";
            if($this->isDrillable($nid) && $this->inContext($context,$nid)){
                $dImage = "<img src=\"".$this->widgetPath."vr";
                if($lastRow){
                    $dImage .= "_l.gif\" border=0 align=top>";
                }else{
                    $dImage .= ".gif\" border=0 align=top>";
                }
                $depth[] = $dImage;
                $output .= $this->drawContextTree($nid,$depth,$context,0);
                $dImage = array_pop($depth);
            }
            unset($children);
        }else{//end isRoot protection
            if($this->gtu_debug)echo "%%% drawContextTree() - NOT Visible or Root ($nid,$pid,$eid,$tid)<br>\n";
        }//end isRoot protection
        next($tree);
    }//end while( each()) 
    if($firstTime)$output .= "</td></tr></table>\n";
    if($this->gtu_debug)echo "%%% drawContextTree() - END<br>\n";
    return $output;
}

function widget($drillable,$drilled,$lastRow){
    if($this->useWidgets){
        $filename = $this->widgetPath;
        if($drillable){
            if($drilled){
                $filename.="minus";
            }else{
                $filename.="plus";
            }
        }else{
            ///not drillable must no plus minus...
            $filename .= "hr";
        }    
        if($lastRow){
            $filename .= "_l.gif";   
        }else{
            $filename .= ".gif";
        }
        $rtn = "<img src=\"".$filename."\" border=0 align=top>";
    }else{
        $rtn="";
    }
    return $rtn;
}

function indent($depth){
    $output = "<tr>\n";
    $output .= "<td nowrap>";
    if($this->useWidgets){
        for($loop=0;$loop<count($depth);$loop++){
            $output .= $depth[$loop];           
        }
    }else{
        //indent the table to the proper depth.
        for($loop=1;$loop<=count($depth);$loop++){
            //$output .= "<code>|</code>";
            $output .= "<code></code>";
            if($loop==count($depth)){
                $output .= "<code>\__</code>";
            }
        
        }
    }
    return $output;
}

function showActions($suffix,$tid){//actions that follow for this type
    $output = "";
// $action is an array of serialized arrays
  $actn = $this->action[$tid];
    $end = count($actn);
    for($loop=0;$loop<$end;$loop++){
        $act = unserialize($actn[$loop]);
        $output .= " <a href=\"".$act[url].$suffix."\" target=\"$act[target]\">$act[text]</a>";
        if($loop != ($end-1)) $output .= " |";
    }
    return $output;
}

function isVisible($tid){
//Visible array defines the type of items to show in the tree
//if the type constant is in the array, all items of that type
//will appear in the tree.  If not, then all items are hidden
    if($this->gtu_debug||0)echo "%%% isVisible() - ".$this->visible[$tid]."(tid=$tid)<br>\n";
    if($this->visible[$tid]==0)return 0;
    return 1;
    
}

function alpha_weight($input,$dbg=0){
    $out = "0.";
    //get string length
    $end = strlen($input);
    //loop though each char, 
    for($i=0;$i<$end;$i++){
        //prefix with zero padding
        $temp="000".ord(substr($input,$i));
        //add last two digits to string starting with "0.0" 
        $out .= substr($temp,-3);       
    }
    //return $out
    if($dbg)echo "input = $input, output = $out<br>\n";
    return $out;
    
}

function sortTree($gid=0){
    //Sorts alpha numerically based on call back function
    $sub_dbg = $this->gtu_debug||false;
    if($sub_dbg)echo "=====> Tree Sort BEGIN (gid = $gid)<br>";
    if($gid==0)$gid=$this->c_rid;
    $func = $this->get_text_func;
    $tree = $this->listNodes($gid);
    if($sub_dbg)echo "Callback Function = $func<br>";
    if($sub_dbg)echo "Branches = ".count($tree)."<br>";
    while(list($nid,$pid,$eid,$tid) = @current($tree)){
        if(!$this->isRoot($nid)){
            if($this->isDrillable($nid)){
                $this->sortTree($nid);
            }
            $this->modNodeWeight($nid,$this->alpha_weight($func($eid,$tid),$sub_dbg));
        }
        @next($tree);
    }
    if($sub_dbg)echo "<===== Tree Sort END<br>";
}


/**********************************************************************
*** Debug Display routines
**********************************************************************/

    function drawTree($gid){
        $tree = $this->listNodes($gid);
        $output = "<ul>";
        while(list($nid,$pid,$eid,$tid) = @current($tree)){
            $output .= "<li>(".$nid.",".$pid.",".$eid.",".$tid.")<br>";
            $children = $this->listNodes($nid);
            if(is_array($children)&&!$this->isRoot($nid)){
                $output .= $this->drawTree($nid);
            }
            next($tree);
            unset($children);
        }
        $output .= "</ul>";
        return $output;
    }

    function showAllGTU(){
       echo "Show All GTU:<p>\n";
       $result = mysql_query("select gid, handle from $this->gtu_table_name"."_reg");
         while(list($gid,$handle) = mysql_fetch_row($result)){
              echo "Tree: ".$handle."<br>\n";
              $this->drawTree($gid);
         }
    }

/*******************************************************************************************
*** Private GTU Context Functions
*******************************************************************************************/
function makeContext($gid){
    $temp = ':';
    $temp .= "$gid";
    $temp .= ':';
    return $temp;
}

function inContext($context,$gid){
    $temp = $this->makeContext($gid);
    if($this->all_in_context)return 1;//everything is in the context.
    if(strstr($context,$temp)){
        return 1;
    }else{
        return 0;
    }
}

function toggleContext($context,$gid){
    if($this->inContext($context,$gid)){
        //remove from context string
            $temp1 = str_replace ($this->makeContext($gid),"", $context);
    }else{
        //add to the end of the context string
            $temp1  = $context;
            $temp1 .= $this->makeContext($gid);
    }
    return $temp1;
}

/*******************************************************************************************
*** State Machine  Functions   COME BACK HERE and do the many state machines....
*******************************************************************************************/
    function serializeStateMachine(){
        $StateMachine["types"]             = serialize($this->types);
        $StateMachine["icons"]             = serialize($this->icons);
        $StateMachine["url"]               = serialize($this->url);
        $StateMachine["target"]         = serialize($this->target);
        $StateMachine["action"]       = serialize($this->action);
        $StateMachine["gtCallBack"] = serialize($this->get_text_func);
        $StateMachine["visible"]         = serialize($this->visible);
        return serialize($StateMachine);
    }

    function unserializeStateMachine($sm){
      $StateMachine = unserialize($sm);
        $this->types                   = unserialize($StateMachine["types"]); 
        $this->icons                   = unserialize($StateMachine["icons"]);
        $this->url                       = unserialize($StateMachine["url"]);
        $this->target                  = unserialize($StateMachine["target"]);
        $this->visible                 = unserialize($StateMachine["visible"]);
        $this->action                  = unserialize($StateMachine["action"]);
        $this->get_text_func     = unserialize($StateMachine["gtCallBack"]);
    }

    function resetStateMachine(){
      $this->types                   = array(); 
        $this->icons                   = array(); 
        $this->url                       = array();
        $this->target                  = array();
        $this->visible                 = array();
        $this->action                  = array();
        $this->get_text_func     = "";
    }
    
    function saveStateMachine(){
        $serSM = $this->serializeStateMachine();
        $root_name = $this->c_root;
        $result = @mysql_query("delete from $this->gtu_table_name"."_sm where id='$root_name'");
        $result = @mysql_query("insert into $this->gtu_table_name"."_sm values('$root_name','$serSM')");
    }

    function restoreStateMachine(){
        $root_name = $this->c_root;
        $result = mysql_query("select handle from $this->gtu_table_name"."_sm where id='$root_name'");
        if(mysql_num_rows($result)>0){
            list($handle) = mysql_fetch_row($result);
            $this->unserializeStateMachine($handle);
        }else{
            $this->resetStateMachine();
        }
    }

/*******************************************************************************************
*** Admin Database Functions
*******************************************************************************************/
    
    
    function setAllRID($gid,$rid){
    //recursive for all drillables...
        $tree = $this->listNodes($gid);
        $cnt = count($tree);
        while(list($nid,$pid,$eid,$tid) = @current($tree)){
            if(!$this->isRoot($nid)&&$this->isDrillable($nid)){
                //recurse
                $this->setAllRID($nid,$rid);
            }
            $this->modNodeRID($nid,$rid);
            next($tree);
        }//end while loop
    }
    
    function updateRIDs(){   //select gid from myh4__myh_gtu_reg;
        $result = mysql_query("select gid from ".$this->gtu_table_name."_reg");
        while(list($gid) = mysql_fetch_row($result)){
            $this->setAllRID($gid,$gid);
        }
    }

    function updateGTUTables(){
        //if this is an older version of GTU update the main table with RID, and weight
        $result = @mysql_query("alter table ".$this->gtu_table_name." add rid int(11) not null");
        $result = @mysql_query("alter table ".$this->gtu_table_name." add weight float default 1.0 not null");
        //populate the rid values...
        $this->updateRIDs();
        //leave the weights as default... let the tree owners do that...        
    }
    
    function createGTUTables(){
    //Master GTU Table
        $cmd =  "CREATE TABLE $this->gtu_table_name (";
        $cmd .= "  gid int(11) DEFAULT '1' NOT NULL auto_increment,";
        $cmd .= "  pid int(11) NOT NULL,";
        $cmd .= "  eid int(11) NOT NULL,";
        $cmd .= "  tid int(11) NOT NULL,";
        $cmd .= "  rid int(11) NOT NULL,";
        $cmd .= "  weight FLOAT DEFAULT 1.0 NOT NULL,";
        $cmd .= "  PRIMARY KEY (gid)";
        $cmd .= ")";
      $err = mysql_query($cmd);
    //GTU Secondary index = PID
        $cmd =  "CREATE INDEX $this->gtu_table_name"."_idx ON $this->gtu_table_name (";
        $cmd .= "  pid";
        $cmd .= ")";
      $err = mysql_query($cmd);
    //GTU Registry (root handle <-> root gid )
        $cmd =  "CREATE TABLE $this->gtu_table_name"."_reg (";
        $cmd .= "  gid int(11) DEFAULT '1' NOT NULL,";
        $cmd .= "  handle varchar(80) DEFAULT '' NOT NULL,";
        $cmd .= "  PRIMARY KEY (gid)";
        $cmd .= ")";
        $err = mysql_query($cmd);
    //GTU State Machine Cache 
        $cmd =  "CREATE TABLE $this->gtu_table_name"."_sm (";
        $cmd .= "  id varchar(80) DEFAULT '' NOT NULL,";
        $cmd .= "  handle blob DEFAULT '' NOT NULL,";
        $cmd .= "  PRIMARY KEY (id)";
        $cmd .= ")";
        $err = mysql_query($cmd);
            
    }

    function dropGTUTables(){
      $err = mysql_query("drop index $this->gtu_table_name"."_idx on $this->gtu_table_name");
      $err = mysql_query("drop table $this->gtu_table_name");
      $err = mysql_query("drop table $this->gtu_table_name"."_reg");
      $err = mysql_query("drop table $this->gtu_table_name"."_sm");
    }

}
?>